@@ -1,4 +1,5 @@
 #!/usr/bin/env  python2
+from __future__ import print_function
 __license__   = 'GPL v3'
 __copyright__ = '2008, Kovid Goyal kovid@kovidgoyal.net'
 __docformat__ = 'restructuredtext en'
@@ -27,6 +28,7 @@ from calibre.utils.config import tweaks
 
 pretty_print_opf = False
 
+
 class PrettyPrint(object):
 
     def __enter__(self):
@@ -36,8 +38,11 @@ class PrettyPrint(object):
     def __exit__(self, *args):
         global pretty_print_opf
         pretty_print_opf = False
+
+
 pretty_print = PrettyPrint()
 
+
 class Resource(object):  # {{{
 
     '''
@@ -121,6 +126,7 @@ class Resource(object):  # {{{
 
 # }}}
 
+
 class ResourceCollection(object):  # {{{
 
     def __init__(self):
@@ -174,6 +180,7 @@ class ResourceCollection(object):  # {{{
 
 # }}}
 
+
 class ManifestItem(Resource):  # {{{
 
     @staticmethod
@@ -190,6 +197,7 @@ class ManifestItem(Resource):  # {{{
     def media_type(self):
         def fget(self):
             return self.mime_type
+
         def fset(self, val):
             self.mime_type = val
         return property(fget=fget, fset=fset)
@@ -212,6 +220,7 @@ class ManifestItem(Resource):  # {{{
 
 # }}}
 
+
 class Manifest(ResourceCollection):  # {{{
 
     def append_from_opf_manifest_item(self, item, dir):
@@ -283,6 +292,7 @@ class Manifest(ResourceCollection):  # {{{
 
 # }}}
 
+
 class Spine(ResourceCollection):  # {{{
 
     class Item(Resource):
@@ -357,6 +367,7 @@ class Spine(ResourceCollection):  # {{{
 
 # }}}
 
+
 class Guide(ResourceCollection):  # {{{
 
     class Reference(Resource):
@@ -395,6 +406,7 @@ class Guide(ResourceCollection):  # {{{
 
 # }}}
 
+
 class MetadataField(object):
 
     def __init__(self, name, is_dc=True, formatter=None, none_is=None,
@@ -437,6 +449,7 @@ class MetadataField(object):
             elem = obj.create_metadata_element(self.name, is_dc=self.is_dc)
         obj.set_text(elem, self.renderer(val))
 
+
 class TitleSortField(MetadataField):
 
     def __get__(self, obj, type=None):
@@ -465,6 +478,7 @@ class TitleSortField(MetadataField):
                     if attr.endswith('file-as'):
                         del match.attrib[attr]
 
+
 def serialize_user_metadata(metadata_elem, all_user_metadata, tail='\n'+(' '*8)):
     from calibre.utils.config import to_json
     from calibre.ebooks.metadata.book.json_codec import (object_to_unicode,
@@ -495,6 +509,7 @@ def dump_dict(cats):
     return json.dumps(object_to_unicode(cats), ensure_ascii=False,
             skipkeys=True)
 
+
 class OPF(object):  # {{{
 
     MIMETYPE         = 'application/oebps-package+xml'
@@ -519,16 +534,16 @@ class OPF(object):  # {{{
     authors_path    = XPath('descendant::*[re:match(name(), "creator", "i") and (@role="aut" or @opf:role="aut" or (not(@role) and not(@opf:role)))]')
     bkp_path        = XPath('descendant::*[re:match(name(), "contributor", "i") and (@role="bkp" or @opf:role="bkp")]')
     tags_path       = XPath('descendant::*[re:match(name(), "subject", "i")]')
-    isbn_path       = XPath('descendant::*[re:match(name(), "identifier", "i") and '+
+    isbn_path       = XPath('descendant::*[re:match(name(), "identifier", "i") and '
                             '(re:match(@scheme, "isbn", "i") or re:match(@opf:scheme, "isbn", "i"))]')
     pubdate_path    = XPath('descendant::*[re:match(name(), "date", "i")]')
-    raster_cover_path = XPath('descendant::*[re:match(name(), "meta", "i") and ' +
+    raster_cover_path = XPath('descendant::*[re:match(name(), "meta", "i") and '
             're:match(@name, "cover", "i") and @content]')
     guide_cover_path = XPath('descendant::*[local-name()="guide"]/*[local-name()="reference" and re:match(@type, "cover", "i")]/@href')
     identifier_path = XPath('descendant::*[re:match(name(), "identifier", "i")]')
-    application_id_path = XPath('descendant::*[re:match(name(), "identifier", "i") and '+
+    application_id_path = XPath('descendant::*[re:match(name(), "identifier", "i") and '
                             '(re:match(@opf:scheme, "calibre|libprs500", "i") or re:match(@scheme, "calibre|libprs500", "i"))]')
-    uuid_id_path    = XPath('descendant::*[re:match(name(), "identifier", "i") and '+
+    uuid_id_path    = XPath('descendant::*[re:match(name(), "identifier", "i") and '
                             '(re:match(@opf:scheme, "uuid", "i") or re:match(@scheme, "uuid", "i"))]')
     languages_path  = XPath('descendant::*[local-name()="language"]')
 
@@ -691,7 +706,7 @@ class OPF(object):  # {{{
     def itermanifest(self):
         return self.manifest_path(self.root)
 
-    def create_manifest_item(self, href, media_type):
+    def create_manifest_item(self, href, media_type, append=False):
         ids = [i.get('id', None) for i in self.itermanifest()]
         id = None
         for c in xrange(1, sys.maxint):
@@ -703,6 +718,9 @@ class OPF(object):  # {{{
         ans = etree.Element('{%s}item'%self.NAMESPACES['opf'],
                              attrib={'id':id, 'href':href, 'media-type':media_type})
         ans.tail = '\n\t\t'
+        if append:
+            manifest = self.manifest_ppath(self.root)[0]
+            manifest.append(ans)
         return ans
 
     def replace_manifest_item(self, item, items):
@@ -714,19 +732,6 @@ class OPF(object):  # {{{
         manifest[index:index+1] = items
         return [i.get('id') for i in items]
 
-    def add_path_to_manifest(self, path, media_type):
-        path = os.path.abspath(path)
-        for i in self.itermanifest():
-            xpath = os.path.join(self.base_dir, *(i.get('href', '').split('/')))
-            if os.path.abspath(xpath) == path:
-                return i.get('id')
-        href = os.path.relpath(path, self.base_dir).replace(os.sep, '/')
-        item = self.create_manifest_item(href, media_type)
-        manifest = self.manifest_ppath(self.root)[0]
-        manifest.append(item)
-        self.manifest.append_from_opf_manifest_item(item, self.basedir)
-        return item.get('id')
-
     def iterspine(self):
         return self.spine_path(self.root)
 
@@ -934,10 +939,24 @@ class OPF(object):  # {{{
                 return self.get_text(match) or None
 
         def fset(self, val):
+            uuid_id = None
+            for attr in self.root.attrib:
+                if attr.endswith('unique-identifier'):
+                    uuid_id = self.root.attrib[attr]
+                    break
+
             matches = self.isbn_path(self.metadata)
             if not val:
                 for x in matches:
-                    x.getparent().remove(x)
+                    xid = x.get('id', None)
+                    is_package_identifier = uuid_id is not None and uuid_id == xid
+                    if is_package_identifier:
+                        self.set_text(x, str(uuid.uuid4()))
+                        for attr in x.attrib:
+                            if attr.endswith('scheme'):
+                                x.attrib[attr] = 'uuid'
+                    else:
+                        x.getparent().remove(x)
                 return
             if not matches:
                 attrib = {'{%s}scheme'%self.NAMESPACES['opf']: 'ISBN'}
@@ -1135,6 +1154,11 @@ class OPF(object):  # {{{
                 if k == 'page-progression-direction' or k.endswith('}page-progression-direction'):
                     return v
 
+    @property
+    def primary_writing_mode(self):
+        for m in self.XPath('//*[local-name()="meta" and @name="primary-writing-mode" and @content]')(self.root):
+            return m.get('content')
+
     def guess_cover(self):
         '''
         Try to guess a cover. Needed for some old/badly formed OPF files.
@@ -1283,6 +1307,7 @@ class OPF(object):  # {{{
             smap[child.get('name')] = (child, self.metadata.index(child))
         if len(smap) == 2 and smap['calibre:series'][1] > smap['calibre:series_index'][1]:
             s, si = smap['calibre:series'][0], smap['calibre:series_index'][0]
+
             def swap(attr):
                 t = s.get(attr, '')
                 s.set(attr, si.get(attr, '')), si.set(attr, t)
@@ -1342,6 +1367,7 @@ class OPF(object):  # {{{
 
 # }}}
 
+
 class OPFCreator(Metadata):
 
     def __init__(self, base_path, other):
@@ -1354,6 +1380,7 @@ class OPFCreator(Metadata):
         Metadata.__init__(self, title='', other=other)
         self.base_path = os.path.abspath(base_path)
         self.page_progression_direction = None
+        self.primary_writing_mode = None
         if self.application_id is None:
             self.application_id = str(uuid.uuid4())
         if not isinstance(self.toc, TOC):
@@ -1516,6 +1543,8 @@ class OPFCreator(Metadata):
             from calibre.ebooks.metadata.book.json_codec import object_to_unicode
             a(CAL_ELEM('calibre:user_categories',
                        json.dumps(object_to_unicode(self.user_categories))))
+        if self.primary_writing_mode:
+            a(M.meta(name='primary-writing-mode', content=self.primary_writing_mode))
         manifest = E.manifest()
         if self.manifest is not None:
             for ref in self.manifest:
@@ -1556,6 +1585,7 @@ class OPFCreator(Metadata):
                 guide
         )
         root.set('unique-identifier', __appname__+'_id')
+        root.set('version', '2.0')
         raw = etree.tostring(root, pretty_print=True, xml_declaration=True,
                 encoding=encoding)
         raw = raw.replace(DNS, OPF2_NS)
@@ -1599,6 +1629,7 @@ def metadata_to_opf(mi, as_string=True, default_lang=None):
     metadata = root[0]
     guide = root[1]
     metadata[0].tail = '\n'+(' '*8)
+
     def factory(tag, text=None, sort=None, role=None, scheme=None, name=None,
             content=None):
         attrib = {}
@@ -1704,7 +1735,7 @@ def test_m2o():
     mi.rights = 'yes'
     mi.cover = os.path.abspath('asd.jpg')
     opf = metadata_to_opf(mi)
-    print opf
+    print(opf)
     newmi = MetaInformation(OPF(StringIO(opf)))
     for attr in ('author_sort', 'title_sort', 'comments',
                     'publisher', 'series', 'series_index', 'rating',
@@ -1714,10 +1745,10 @@ def test_m2o():
                     'pubdate', 'rights', 'publication_type'):
         o, n = getattr(mi, attr), getattr(newmi, attr)
         if o != n and o.strip() != n.strip():
-            print 'FAILED:', attr, getattr(mi, attr), '!=', getattr(newmi, attr)
+            print('FAILED:', attr, getattr(mi, attr), '!=', getattr(newmi, attr))
     if mi.get_identifiers() != newmi.get_identifiers():
-        print 'FAILED:', 'identifiers', mi.get_identifiers(),
-        print '!=', newmi.get_identifiers()
+        print('FAILED:', 'identifiers', mi.get_identifiers(), end=' ')
+        print('!=', newmi.get_identifiers())
 
 
 class OPFTest(unittest.TestCase):
@@ -1786,12 +1817,15 @@ class OPFTest(unittest.TestCase):
         self.opf.smart_update(MetaInformation(self.opf))
         self.testReading()
 
+
 def suite():
     return unittest.TestLoader().loadTestsFromTestCase(OPFTest)
 
+
 def test():
     unittest.TextTestRunner(verbosity=2).run(suite())
 
+
 def test_user_metadata():
     from cStringIO import StringIO
     mi = Metadata('Test title', ['test author1', 'test author2'])
@@ -1815,7 +1849,8 @@ def test_user_metadata():
     opf2 = OPF(f2)
     assert um == opf._user_metadata_
     assert um == opf2._user_metadata_
-    print opf.render()
+    print(opf.render())
+
 
 if __name__ == '__main__':
     # test_user_metadata()
