@@ -1,32 +1,46 @@
 __license__   = 'GPL v3'
 __copyright__ = '2008, Kovid Goyal <kovid at kovidgoyal.net>'
 """ The GUI """
-import os, sys, Queue, threading, glob, signal
+import glob
+import os
+import Queue
+import signal
+import sys
+import threading
 from contextlib import contextmanager
-from threading import RLock, Lock
-from PyQt5.QtWidgets import QStyle  # Gives a nicer error message than import from Qt
-from PyQt5.Qt import (
-    QFileInfo, QObject, QBuffer, Qt, QByteArray, QTranslator, QSocketNotifier,
-    QCoreApplication, QThread, QEvent, QTimer, pyqtSignal, QDateTime, QFontMetrics,
-    QDesktopServices, QFileDialog, QFileIconProvider, QSettings, QIcon, QStringListModel,
-    QApplication, QDialog, QUrl, QFont, QFontDatabase, QLocale, QFontInfo)
+from threading import Lock, RLock
 
-from calibre import prints
-from calibre.constants import (islinux, iswindows, isbsd, isfrozen, isosx, is_running_from_develop,
-        plugins, config_dir, filesystem_encoding, isxp, DEBUG, __version__, __appname__ as APP_UID)
-from calibre.ptempfile import base_dir
-from calibre.gui2.linux_file_dialogs import check_for_linux_native_dialogs, linux_native_dialog
-from calibre.gui2.qt_file_dialogs import FileDialog
-from calibre.utils.config import Config, ConfigProxy, dynamic, JSONConfig
+from PyQt5.Qt import (
+    QT_VERSION, QApplication, QBuffer, QByteArray, QCoreApplication, QDateTime,
+    QDesktopServices, QDialog, QEvent, QFileDialog, QFileIconProvider, QFileInfo,
+    QFont, QFontDatabase, QFontInfo, QFontMetrics, QIcon, QLocale, QObject,
+    QSettings, QSocketNotifier, QStringListModel, Qt, QThread, QTimer, QTranslator,
+    QUrl, pyqtSignal
+)
+from PyQt5.QtWidgets import QStyle  # Gives a nicer error message than import from Qt
+
+from calibre import as_unicode, prints
+from calibre.constants import (
+    DEBUG, __appname__ as APP_UID, __version__, config_dir, filesystem_encoding,
+    is_running_from_develop, isbsd, isfrozen, islinux, isosx, iswindows, isxp,
+    plugins
+)
 from calibre.ebooks.metadata import MetaInformation
+from calibre.gui2.linux_file_dialogs import (
+    check_for_linux_native_dialogs, linux_native_dialog
+)
+from calibre.gui2.qt_file_dialogs import FileDialog
+from calibre.ptempfile import base_dir
+from calibre.utils.config import Config, ConfigProxy, JSONConfig, dynamic
 from calibre.utils.date import UNDEFINED_DATE
-from calibre.utils.localization import get_lang
 from calibre.utils.file_type_icons import EXT_MAP
+from calibre.utils.localization import get_lang
 
 try:
     NO_URL_FORMATTING = QUrl.None_
 except AttributeError:
-    NO_URL_FORMATTING = QUrl.None
+    NO_URL_FORMATTING = getattr(QUrl, 'None')
+
 
 # Setup gprefs {{{
 gprefs = JSONConfig('gui')
@@ -87,6 +101,13 @@ def create_defs():
             'Similar Books', 'Tweak ePub', None, 'Remove Books',
             )
 
+    defs['action-layout-context-menu-split'] = (
+            'Edit Metadata', 'Send To Device', 'Save To Disk',
+            'Connect Share', 'Copy To Library', None,
+            'Convert Books', 'View', 'Open Folder', 'Show Book Details',
+            'Similar Books', 'Tweak ePub', None, 'Remove Books',
+            )
+
     defs['action-layout-context-menu-device'] = (
             'View', 'Save To Disk', None, 'Remove Books', None,
             'Add To Library', 'Edit Collections', 'Match Books'
@@ -139,6 +160,7 @@ def create_defs():
     defs['cover_grid_show_title'] = False
     defs['cover_grid_texture'] = None
     defs['show_vl_tabs'] = False
+    defs['vl_tabs_closable'] = True
     defs['show_highlight_toggle_button'] = False
     defs['add_comments_to_email'] = False
     defs['cb_preserve_aspect_ratio'] = False
@@ -148,6 +170,7 @@ def create_defs():
     defs['emblem_position'] = 'left'
     defs['metadata_diff_mark_rejected'] = False
     defs['tag_browser_show_counts'] = True
+    defs['tag_browser_show_tooltips'] = True
     defs['row_numbers_in_book_list'] = True
     defs['hidpi'] = 'auto'
     defs['tag_browser_item_padding'] = 0.5
@@ -156,6 +179,11 @@ def create_defs():
     defs['qv_dclick_changes_column'] = True
     defs['qv_retkey_changes_column'] = True
     defs['qv_follows_column'] = False
+    defs['book_details_narrow_comments_layout'] = 'float'
+    defs['book_list_split'] = False
+    defs['wrap_toolbar_text'] = False
+    defs['dnd_merge'] = True
+    defs['booklist_grid'] = False
 
 
 create_defs()
@@ -337,9 +365,9 @@ def extension(path):
 def warning_dialog(parent, title, msg, det_msg='', show=False,
         show_copy_button=True):
     from calibre.gui2.dialogs.message_box import MessageBox
-    d = MessageBox(MessageBox.WARNING, _('WARNING:')+ ' ' +
-            title, msg, det_msg, parent=parent,
-            show_copy_button=show_copy_button)
+    d = MessageBox(MessageBox.WARNING, _('WARNING:'
+        )+ ' ' + title, msg, det_msg, parent=parent,
+        show_copy_button=show_copy_button)
     if show:
         return d.exec_()
     return d
@@ -348,9 +376,9 @@ def warning_dialog(parent, title, msg, det_msg='', show=False,
 def error_dialog(parent, title, msg, det_msg='', show=False,
         show_copy_button=True):
     from calibre.gui2.dialogs.message_box import MessageBox
-    d = MessageBox(MessageBox.ERROR, _('ERROR:')+ ' ' +
-            title, msg, det_msg, parent=parent,
-                    show_copy_button=show_copy_button)
+    d = MessageBox(MessageBox.ERROR, _('ERROR:'
+        ) + ' ' + title, msg, det_msg, parent=parent,
+        show_copy_button=show_copy_button)
     if show:
         return d.exec_()
     return d
@@ -784,6 +812,8 @@ class Application(QApplication):
 
     def __init__(self, args, force_calibre_style=False, override_program_name=None, headless=False, color_prefs=gprefs):
         self.file_event_hook = None
+        if isfrozen and QT_VERSION <= 0x050700 and 'wayland' in os.environ.get('QT_QPA_PLATFORM', ''):
+            os.environ['QT_QPA_PLATFORM'] = 'xcb'
         if override_program_name:
             args = [override_program_name] + args[1:]
         if headless:
@@ -800,6 +830,8 @@ class Application(QApplication):
         QApplication.setOrganizationDomain(QApplication.organizationName())
         QApplication.setApplicationVersion(__version__)
         QApplication.setApplicationName(APP_UID)
+        if override_program_name and hasattr(QApplication, 'setDesktopFileName'):
+            QApplication.setDesktopFileName(override_program_name)
         QApplication.__init__(self, qargs)
         self.setAttribute(Qt.AA_UseHighDpiPixmaps)
         self.setAttribute(Qt.AA_SynthesizeTouchForUnhandledMouseEvents, False)
@@ -831,6 +863,11 @@ class Application(QApplication):
             if s is not None:
                 font.setStretch(s)
             QApplication.setFont(font)
+        if not isosx and not iswindows:
+            # Qt 5.10.1 on Linux resets the global font on first event loop tick.
+            # So workaround it by setting the font once again in a timer.
+            font_from_prefs = self.font()
+            QTimer.singleShot(0, lambda : QApplication.setFont(font_from_prefs))
         self.line_height = max(12, QFontMetrics(self.font()).lineSpacing())
 
         dl = QLocale(get_lang())
@@ -872,6 +909,14 @@ class Application(QApplication):
             # Qt 5 bug: https://bugreports.qt-project.org/browse/QTBUG-41125
             self.aboutToQuit.connect(self.flush_clipboard)
 
+        if isosx:
+            cocoa, err = plugins['cocoa']
+            if err:
+                raise RuntimeError('Failed to load cocoa plugin with error: {}'.format(err))
+            cft = cocoa.cursor_blink_time()
+            if cft >= 0:
+                self.setCursorFlashTime(int(cft))
+
     def setup_ui_font(self):
         f = QFont(QApplication.font())
         q = (f.family(), f.pointSize())
@@ -910,7 +955,7 @@ class Application(QApplication):
         if iswindows or isosx:
             using_calibre_style = gprefs['ui_style'] != 'system'
         else:
-            using_calibre_style = 'CALIBRE_USE_SYSTEM_THEME' not in os.environ
+            using_calibre_style = os.environ.get('CALIBRE_USE_SYSTEM_THEME', '0') == '0'
         if force_calibre_style:
             using_calibre_style = True
         self.using_calibre_style = using_calibre_style
@@ -938,6 +983,7 @@ class Application(QApplication):
             'MessageBoxCritical': u'dialog_error.png',
             'MessageBoxQuestion': u'dialog_question.png',
             'BrowserReload': u'view-refresh.png',
+            'LineEditClearButton': u'clear_left.png',
         }.iteritems():
             if v not in pcache:
                 p = I(v)
@@ -1005,11 +1051,15 @@ class Application(QApplication):
 
     def setup_unix_signals(self):
         import fcntl
-        read_fd, write_fd = os.pipe()
-        cloexec_flag = getattr(fcntl, 'FD_CLOEXEC', 1)
-        for fd in (read_fd, write_fd):
-            flags = fcntl.fcntl(fd, fcntl.F_GETFD)
-            fcntl.fcntl(fd, fcntl.F_SETFD, flags | cloexec_flag | os.O_NONBLOCK)
+        if hasattr(os, 'pipe2'):
+            read_fd, write_fd = os.pipe2(os.O_CLOEXEC | os.O_NONBLOCK)
+        else:
+            read_fd, write_fd = os.pipe()
+            cloexec_flag = getattr(fcntl, 'FD_CLOEXEC', 1)
+            for fd in (read_fd, write_fd):
+                flags = fcntl.fcntl(fd, fcntl.F_GETFD)
+                fcntl.fcntl(fd, fcntl.F_SETFD, flags | cloexec_flag | os.O_NONBLOCK)
+
         for sig in (signal.SIGINT, signal.SIGTERM):
             signal.signal(sig, lambda x, y: None)
             signal.siginterrupt(sig, False)
@@ -1285,3 +1335,50 @@ def secure_web_page(qwebpage_or_qwebsettings):
 
 empty_model = QStringListModel([''])
 empty_index = empty_model.index(0)
+
+
+def get_app_uid():
+    import ctypes
+    from ctypes import wintypes
+    lpBuffer = wintypes.LPWSTR()
+    try:
+        AppUserModelID = ctypes.windll.shell32.GetCurrentProcessExplicitAppUserModelID
+    except Exception:  # Vista has no app uids
+        return
+    AppUserModelID.argtypes = [wintypes.LPWSTR]
+    AppUserModelID.restype = wintypes.HRESULT
+    try:
+        AppUserModelID(ctypes.cast(ctypes.byref(lpBuffer), wintypes.LPWSTR))
+    except Exception:
+        return
+    appid = lpBuffer.value
+    ctypes.windll.ole32.CoTaskMemFree(lpBuffer)
+    return appid
+
+
+def set_app_uid(val):
+    import ctypes
+    from ctypes import wintypes
+    try:
+        AppUserModelID = ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID
+    except Exception:  # Vista has no app uids
+        return False
+    AppUserModelID.argtypes = [wintypes.LPCWSTR]
+    AppUserModelID.restype = wintypes.HRESULT
+    try:
+        AppUserModelID(unicode(val))
+    except Exception as err:
+        prints(u'Failed to set app uid with error:', as_unicode(err))
+        return False
+    return True
+
+
+def add_to_recent_docs(path):
+    from win32com.shell import shell, shellcon
+    path = unicode(path)
+    app_id = get_app_uid()
+    if app_id is None:
+        shell.SHAddToRecentDocs(shellcon.SHARD_PATHW, path)
+    else:
+        item = shell.SHCreateItemFromParsingName(path, None, shell.IID_IShellItem)
+        shell.SHAddToRecentDocs(shellcon.SHARD_APPIDINFO, (item, app_id))
